@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2020, Sonos, Inc.
+ * Copyright (c) 2014-2016, Sonos, Inc.
  *
  * SPDX-License-Identifier:	GPL-2.0
  */
@@ -9,7 +9,6 @@
 #include <linux/errno.h>
 #include <linux/version.h>
 
-//ALLEN: All the functions are copied from new kernels, so that old kernels can use it.
 
 #if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,22)
 
@@ -86,7 +85,7 @@ static int xlate_proc_name_ex(const char *name,
 	struct proc_dir_entry	*de;
 	int			len;
 
-	de = *ret;  // start at *ret not proc_root
+	de = *ret;
 	if (!de)
 		de = &proc_root;
 
@@ -119,14 +118,11 @@ static struct proc_dir_entry *__proc_create(
 	const char *fn = name;
 	int len;
 
-	/* make sure name is valid */
 	if (!name || !strlen(name)) goto out;
 
-	// Call the extrapolated version which keeps the parent directory
 	if (xlate_proc_name_ex(name, parent, &fn) != 0)
 		goto out;
 
-	/* At this point there must not be any '/' characters beyond *fn */
 	if (strchr(fn, '/'))
 		goto out;
 
@@ -185,7 +181,7 @@ EXPORT_SYMBOL(proc_create_data);
 
 #endif
 
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,39)
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,35)
 static inline void *__PDE_DATA(const struct inode *inode)
 {
 	return PDE(inode)->data;
