@@ -13,7 +13,7 @@
 #include <linux/proc_fs.h>
 #include <linux/seq_file.h>
 
-#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,39)
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,35)
 	void *PDE_DATA(const struct inode *);
 	static inline struct inode *file_inode(struct file *f)
 	{
@@ -50,7 +50,6 @@
 	int single_release(struct inode *in_inode, struct file *in_file);
 #endif
 
-//ALLEN: Following Macros used by all kernels
 
 #define SONOS_PROCFS_PERM_READ	(S_IRUSR | S_IRGRP | S_IROTH)
 #define SONOS_PROCFS_PERM_WRITE	(S_IWUSR | S_IWGRP | S_IWOTH)
@@ -73,4 +72,4 @@
 #define SONOS_PROCFS_INIT(procReadFunc, procWriteFunc, fopsStr)							\
     SONOS_PROCFS_INIT_WITH_SIZE(procReadFunc, procWriteFunc, fopsStr, PAGE_SIZE)
 
-#endif // ifndef _SONOS_PROCFS_H
+#endif
