@@ -6,9 +6,11 @@
  *  GK 2/5/95  -  Changed to support mounting root fs via NFS
  *  Added initrd & change_root: Werner Almesberger & Hans Lermen, Feb '96
  *  Moan early if gcc is old, avoiding bogus kernels - Paul Gortmaker, May '96
- *  Simplified starting of init:  Michael A. Griffith <grif@acm.org>
+ *  Simplified starting of init:  Michael A. Griffith <grif@acm.org> 
  */
 
+#define DEBUG		/* Enable initcall_debug */
+
 #include <linux/types.h>
 #include <linux/module.h>
 #include <linux/proc_fs.h>
@@ -20,7 +22,6 @@
 #include <linux/delay.h>
 #include <linux/ioport.h>
 #include <linux/init.h>
-#include <linux/smp_lock.h>
 #include <linux/initrd.h>
 #include <linux/bootmem.h>
 #include <linux/acpi.h>
@@ -32,7 +33,6 @@
 #include <linux/start_kernel.h>
 #include <linux/security.h>
 #include <linux/smp.h>
-#include <linux/workqueue.h>
 #include <linux/profile.h>
 #include <linux/rcupdate.h>
 #include <linux/moduleparam.h>
@@ -66,11 +66,15 @@
 #include <linux/ftrace.h>
 #include <linux/async.h>
 #include <linux/kmemcheck.h>
-#include <linux/kmemtrace.h>
 #include <linux/sfi.h>
 #include <linux/shmem_fs.h>
 #include <linux/slab.h>
-#include <trace/boot.h>
+#include <linux/perf_event.h>
+#include <linux/file.h>
+#include <linux/ptrace.h>
+#include <linux/blkdev.h>
+#include <linux/elevator.h>
+#include <linux/random.h>
 
 #include <asm/io.h>
 #include <asm/bugs.h>
@@ -82,23 +86,13 @@
 #include <asm/smp.h>
 #endif
 
-#ifdef CONFIG_SONOS_FENWAY
-#include "asm/mpc83xx-gpio.h"
-#include "asm/fenway-gpio.h"
-#endif
-
-#include "mdp.h"
-extern struct manufacturing_data_page sys_mdp;
-
 static int kernel_init(void *);
 
 extern void init_IRQ(void);
 extern void fork_init(unsigned long);
 extern void mca_init(void);
 extern void sbus_init(void);
-extern void prio_tree_init(void);
 extern void radix_tree_init(void);
-extern void free_initmem(void);
 #ifndef CONFIG_DEBUG_RODATA
 static inline void mark_rodata_ro(void) { }
 #endif
@@ -107,15 +101,18 @@ static inline void mark_rodata_ro(void) { }
 extern void tc_init(void);
 #endif
 
-#ifdef CONFIG_SONOS_FENWAY
-#define LED_WHITE          0x01
-#define LED_GREEN          0x02
-#define LED_AMBER          0x04
-#define LED_RED            0x08
-#define LED_GREEN_BUTTON   0x10
-#define LED_OVERRIDE       0x20
-extern int fenway_submodel;
-#endif	// CONFIG_SONOS_FENWAY
+#ifdef CONFIG_SONOS_SECBOOT
+extern void check_sysrq_authorization(void);
+#endif
+
+/*
+ * Debug helper: via this flag we know that we are in 'early bootup code'
+ * where only the boot processor is running with IRQ disabled.  This means
+ * two things - IRQ must not be enabled before the flag is cleared and some
+ * operations which are not allowed with IRQ disabled are allowed while the
+ * flag is set.
+ */
+bool early_boot_irqs_disabled __read_mostly;
 
 enum system_states system_state __read_mostly;
 EXPORT_SYMBOL(system_state);
@@ -141,63 +138,6 @@ static char *static_command_line;
 static char *execute_command;
 static char *ramdisk_execute_command;
 
-#ifdef CONFIG_SMP
-/* Setup configured maximum number of CPUs to activate */
-unsigned int setup_max_cpus = NR_CPUS;
-EXPORT_SYMBOL(setup_max_cpus);
-
-
-/*
- * Setup routine for controlling SMP activation
- *
- * Command-line option of "nosmp" or "maxcpus=0" will disable SMP
- * activation entirely (the MPS table probe still happens, though).
- *
- * Command-line option of "maxcpus=<NUM>", where <NUM> is an integer
- * greater than 0, limits the maximum number of CPUs activated in
- * SMP mode to <NUM>.
- */
-
-void __weak arch_disable_smp_support(void) { }
-
-static int __init nosmp(char *str)
-{
-	setup_max_cpus = 0;
-	arch_disable_smp_support();
-
-	return 0;
-}
-
-early_param("nosmp", nosmp);
-
-/* this is hard limit */
-static int __init nrcpus(char *str)
-{
-	int nr_cpus;
-
-	get_option(&str, &nr_cpus);
-	if (nr_cpus > 0 && nr_cpus < nr_cpu_ids)
-		nr_cpu_ids = nr_cpus;
-
-	return 0;
-}
-
-early_param("nr_cpus", nrcpus);
-
-static int __init maxcpus(char *str)
-{
-	get_option(&str, &setup_max_cpus);
-	if (setup_max_cpus == 0)
-		arch_disable_smp_support();
-
-	return 0;
-}
-
-early_param("maxcpus", maxcpus);
-#else
-static const unsigned int setup_max_cpus = NR_CPUS;
-#endif
-
 /*
  * If set, this is an indication to the drivers that reset the underlying
  * device before going ahead with the initialization otherwise driver might
@@ -218,21 +158,21 @@ static int __init set_reset_devices(char *str)
 
 __setup("reset_devices", set_reset_devices);
 
-static char * argv_init[MAX_INIT_ARGS+2] = { "init", NULL, };
-char * envp_init[MAX_INIT_ENVS+2] = { "HOME=/", "TERM=linux", NULL, };
+static const char * argv_init[MAX_INIT_ARGS+2] = { "init", NULL, };
+const char * envp_init[MAX_INIT_ENVS+2] = { "HOME=/", "TERM=linux", NULL, };
 static const char *panic_later, *panic_param;
 
-extern struct obs_kernel_param __setup_start[], __setup_end[];
+extern const struct obs_kernel_param __setup_start[], __setup_end[];
 
 static int __init obsolete_checksetup(char *line)
 {
-	struct obs_kernel_param *p;
+	const struct obs_kernel_param *p;
 	int had_early_param = 0;
 
 	p = __setup_start;
 	do {
 		int n = strlen(p->str);
-		if (!strncmp(line, p->str, n)) {
+		if (parameqn(line, p->str, n)) {
 			if (p->early) {
 				/* Already done in parse_early_param?
 				 * (Needs exact match on param part).
@@ -241,8 +181,8 @@ static int __init obsolete_checksetup(char *line)
 				if (line[n] == '\0' || line[n] == '=')
 					had_early_param = 1;
 			} else if (!p->setup_func) {
-				printk(KERN_WARNING "Parameter %s is obsolete,"
-				       " ignored\n", p->str);
+				pr_warn("Parameter %s is obsolete, ignored\n",
+					p->str);
 				return 1;
 			} else if (p->setup_func(line + n))
 				return 1;
@@ -278,19 +218,26 @@ early_param("quiet", quiet_kernel);
 
 static int __init loglevel(char *str)
 {
-	get_option(&str, &console_loglevel);
-	return 0;
+	int newlevel;
+
+	/*
+	 * Only update loglevel value when a correct setting was passed,
+	 * to prevent blind crashes (when loglevel being set to 0) that
+	 * are quite hard to debug
+	 */
+	if (get_option(&str, &newlevel)) {
+		console_loglevel = newlevel;
+		return 0;
+	}
+
+	return -EINVAL;
 }
 
 early_param("loglevel", loglevel);
 
-/*
- * Unknown boot options get handed to init, unless they look like
- * unused parameters (modprobe will find them in /proc/cmdline).
- */
-static int __init unknown_bootoption(char *param, char *val)
+/* Change NUL term back to "=", to make "param" the whole string. */
+static int __init repair_env_string(char *param, char *val, const char *unused)
 {
-	/* Change NUL term back to "=", to make "param" the whole string. */
 	if (val) {
 		/* param=val or param="val"? */
 		if (val == param+strlen(param)+1)
@@ -302,6 +249,16 @@ static int __init unknown_bootoption(char *param, char *val)
 		} else
 			BUG();
 	}
+	return 0;
+}
+
+/*
+ * Unknown boot options get handed to init, unless they look like
+ * unused parameters (modprobe will find them in /proc/cmdline).
+ */
+static int __init unknown_bootoption(char *param, char *val, const char *unused)
+{
+	repair_env_string(param, val, unused);
 
 	/* Handle obsolete-style parameters */
 	if (obsolete_checksetup(param))
@@ -340,10 +297,6 @@ static int __init unknown_bootoption(char *param, char *val)
 	return 0;
 }
 
-#ifdef CONFIG_DEBUG_PAGEALLOC
-int __read_mostly debug_pagealloc_enabled = 0;
-#endif
-
 static int __init init_setup(char *str)
 {
 	unsigned int i;
@@ -374,7 +327,7 @@ static int __init rdinit_setup(char *str)
 __setup("rdinit=", rdinit_setup);
 
 #ifndef CONFIG_SMP
-
+static const unsigned int setup_max_cpus = NR_CPUS;
 #ifdef CONFIG_X86_LOCAL_APIC
 static void __init smp_init(void)
 {
@@ -386,37 +339,6 @@ static void __init smp_init(void)
 
 static inline void setup_nr_cpu_ids(void) { }
 static inline void smp_prepare_cpus(unsigned int maxcpus) { }
-
-#else
-
-/* Setup number of possible processor ids */
-int nr_cpu_ids __read_mostly = NR_CPUS;
-EXPORT_SYMBOL(nr_cpu_ids);
-
-/* An arch may set nr_cpu_ids earlier if needed, so this would be redundant */
-static void __init setup_nr_cpu_ids(void)
-{
-	nr_cpu_ids = find_last_bit(cpumask_bits(cpu_possible_mask),NR_CPUS) + 1;
-}
-
-/* Called by boot processor to activate the rest. */
-static void __init smp_init(void)
-{
-	unsigned int cpu;
-
-	/* FIXME: This should be done in userspace --RR */
-	for_each_present_cpu(cpu) {
-		if (num_online_cpus() >= setup_max_cpus)
-			break;
-		if (!cpu_online(cpu))
-			cpu_up(cpu);
-	}
-
-	/* Any cleanup work */
-	printk(KERN_INFO "Brought up %ld CPUs\n", (long)num_online_cpus());
-	smp_cpus_done(setup_max_cpus);
-}
-
 #endif
 
 /*
@@ -445,7 +367,6 @@ static void __init setup_command_line(char *command_line)
 static __initdata DECLARE_COMPLETION(kthreadd_done);
 
 static noinline void __init_refok rest_init(void)
-	__releases(kernel_lock)
 {
 	int pid;
 
@@ -462,34 +383,29 @@ static noinline void __init_refok rest_init(void)
 	kthreadd_task = find_task_by_pid_ns(pid, &init_pid_ns);
 	rcu_read_unlock();
 	complete(&kthreadd_done);
-	unlock_kernel();
 
 	/*
 	 * The boot idle thread must execute schedule()
 	 * at least once to get things moving:
 	 */
 	init_idle_bootup_task(current);
-	preempt_enable_no_resched();
-	schedule();
-	preempt_disable();
-
+	schedule_preempt_disabled();
 	/* Call into cpu_idle with preempt disabled */
-	cpu_idle();
+	cpu_startup_entry(CPUHP_ONLINE);
 }
 
 /* Check for early params. */
-static int __init do_early_param(char *param, char *val)
+static int __init do_early_param(char *param, char *val, const char *unused)
 {
-	struct obs_kernel_param *p;
+	const struct obs_kernel_param *p;
 
 	for (p = __setup_start; p < __setup_end; p++) {
-		if ((p->early && strcmp(param, p->str) == 0) ||
+		if ((p->early && parameq(param, p->str)) ||
 		    (strcmp(param, "console") == 0 &&
 		     strcmp(p->str, "earlycon") == 0)
 		) {
 			if (p->setup_func(val) != 0)
-				printk(KERN_WARNING
-				       "Malformed early option '%s'\n", param);
+				pr_warn("Malformed early option '%s'\n", param);
 		}
 	}
 	/* We accept everything at this stage. */
@@ -498,7 +414,7 @@ static int __init do_early_param(char *param, char *val)
 
 void __init parse_early_options(char *cmdline)
 {
-	parse_args("early options", cmdline, NULL, 0, do_early_param);
+	parse_args("early options", cmdline, NULL, 0, 0, 0, do_early_param);
 }
 
 /* Arch code calls this early on, or if not, just before other parsing. */
@@ -534,9 +450,11 @@ void __init __weak smp_setup_processor_id(void)
 {
 }
 
+# if THREAD_SIZE >= PAGE_SIZE
 void __init __weak thread_info_cache_init(void)
 {
 }
+#endif
 
 /*
  * Set up kernel memory allocators
@@ -544,12 +462,13 @@ void __init __weak thread_info_cache_init(void)
 static void __init mm_init(void)
 {
 	/*
-	 * page_cgroup requires countinous pages as memmap
-	 * and it's bigger than MAX_ORDER unless SPARSEMEM.
+	 * page_cgroup requires contiguous pages,
+	 * bigger than MAX_ORDER unless SPARSEMEM.
 	 */
 	page_cgroup_init_flatmem();
 	mem_init();
 	kmem_cache_init();
+	percpu_init_late();
 	pgtable_cache_init();
 	vmalloc_init();
 }
@@ -557,15 +476,14 @@ static void __init mm_init(void)
 asmlinkage void __init start_kernel(void)
 {
 	char * command_line;
-	extern struct kernel_param __start___param[], __stop___param[];
-
-	smp_setup_processor_id();
+	extern const struct kernel_param __start___param[], __stop___param[];
 
 	/*
 	 * Need to run as early as possible, to initialize the
 	 * lockdep hash:
 	 */
 	lockdep_init();
+	smp_setup_processor_id();
 	debug_objects_early_init();
 
 	/*
@@ -576,42 +494,45 @@ asmlinkage void __init start_kernel(void)
 	cgroup_init_early();
 
 	local_irq_disable();
-	early_boot_irqs_off();
-	early_init_irq_lock_class();
+	early_boot_irqs_disabled = true;
 
 /*
  * Interrupts are still disabled. Do necessary setups, then
  * enable them
  */
-	lock_kernel();
-	tick_init();
 	boot_cpu_init();
 	page_address_init();
-	printk(KERN_NOTICE "%s", linux_banner);
+	pr_notice("%s", linux_banner);
 	setup_arch(&command_line);
 	mm_init_owner(&init_mm, &init_task);
+	mm_init_cpumask(&init_mm);
 	setup_command_line(command_line);
 	setup_nr_cpu_ids();
 	setup_per_cpu_areas();
 	smp_prepare_boot_cpu();	/* arch-specific boot-cpu hooks */
 
-	build_all_zonelists(NULL);
+	build_all_zonelists(NULL, NULL);
 	page_alloc_init();
 
-	printk(KERN_NOTICE "Kernel command line: %s\n", boot_command_line);
+	pr_notice("Kernel command line: %s\n", boot_command_line);
 	parse_early_param();
 	parse_args("Booting kernel", static_command_line, __start___param,
 		   __stop___param - __start___param,
-		   &unknown_bootoption);
+		   -1, -1, &unknown_bootoption);
+
+	jump_label_init();
+
 	/*
 	 * These use large bootmem allocations and must precede
 	 * kmem_cache_init()
 	 */
+	setup_log_buf(0);
 	pidhash_init();
 	vfs_caches_init_early();
 	sort_main_extable();
 	trap_init();
 	mm_init();
+
 	/*
 	 * Set up the scheduler prior starting any interrupts (such as the
 	 * timer interrupt). Full topology setup happens at smp_init()
@@ -623,32 +544,28 @@ asmlinkage void __init start_kernel(void)
 	 * fragile until we cpu_idle() for the first time.
 	 */
 	preempt_disable();
-	if (!irqs_disabled()) {
-		printk(KERN_WARNING "start_kernel(): bug: interrupts were "
-				"enabled *very* early, fixing it\n");
+	if (WARN(!irqs_disabled(), "Interrupts were enabled *very* early, fixing it\n"))
 		local_irq_disable();
-	}
+	idr_init_cache();
+	perf_event_init();
 	rcu_init();
+	tick_nohz_init();
 	radix_tree_init();
 	/* init some links before init_ISA_irqs() */
 	early_irq_init();
 	init_IRQ();
-	prio_tree_init();
+	tick_init();
 	init_timers();
 	hrtimers_init();
 	softirq_init();
 	timekeeping_init();
 	time_init();
 	profile_init();
-	if (!irqs_disabled())
-		printk(KERN_CRIT "start_kernel(): bug: interrupts were "
-				 "enabled early\n");
-	early_boot_irqs_on();
+	call_function_init();
+	WARN(!irqs_disabled(), "Interrupts were enabled early\n");
+	early_boot_irqs_disabled = false;
 	local_irq_enable();
 
-	/* Interrupts are enabled now so all GFP allocations are safe. */
-	gfp_allowed_mask = __GFP_BITS_MASK;
-
 	kmem_cache_init_late();
 
 	/*
@@ -672,19 +589,15 @@ asmlinkage void __init start_kernel(void)
 #ifdef CONFIG_BLK_DEV_INITRD
 	if (initrd_start && !initrd_below_start_ok &&
 	    page_to_pfn(virt_to_page((void *)initrd_start)) < min_low_pfn) {
-		printk(KERN_CRIT "initrd overwritten (0x%08lx < 0x%08lx) - "
-		    "disabling it.\n",
+		pr_crit("initrd overwritten (0x%08lx < 0x%08lx) - disabling it.\n",
 		    page_to_pfn(virt_to_page((void *)initrd_start)),
 		    min_low_pfn);
 		initrd_start = 0;
 	}
 #endif
 	page_cgroup_init();
-	enable_debug_pagealloc();
-	kmemtrace_init();
-	kmemleak_init();
 	debug_objects_mem_init();
-	idr_init_cache();
+	kmemleak_init();
 	setup_per_cpu_pageset();
 	numa_policy_init();
 	if (late_time_init)
@@ -694,8 +607,12 @@ asmlinkage void __init start_kernel(void)
 	pidmap_init();
 	anon_vma_init();
 #ifdef CONFIG_X86
-	if (efi_enabled)
+	if (efi_enabled(EFI_RUNTIME_SERVICES))
 		efi_enter_virtual_mode();
+#endif
+#ifdef CONFIG_X86_ESPFIX64
+	/* Should be run before the first non-init thread is created */
+	init_espfix_bsp();
 #endif
 	thread_info_cache_init();
 	cred_init();
@@ -722,6 +639,11 @@ asmlinkage void __init start_kernel(void)
 	acpi_early_init(); /* before LAPIC and SMP init */
 	sfi_init_late();
 
+	if (efi_enabled(EFI_RUNTIME_SERVICES)) {
+		efi_late_init();
+		efi_free_boot_services();
+	}
+
 	ftrace_init();
 
 	/* Do the rest non-__init'ed, we're now alive */
@@ -739,70 +661,112 @@ static void __init do_ctors(void)
 #endif
 }
 
-int initcall_debug;
+bool initcall_debug;
 core_param(initcall_debug, initcall_debug, bool, 0644);
 
 static char msgbuf[64];
-static struct boot_trace_call call;
-static struct boot_trace_ret ret;
 
-int do_one_initcall(initcall_t fn)
+static int __init_or_module do_one_initcall_debug(initcall_t fn)
+{
+	ktime_t calltime, delta, rettime;
+	unsigned long long duration;
+	int ret;
+
+	pr_debug("calling  %pF @ %i\n", fn, task_pid_nr(current));
+	calltime = ktime_get();
+	ret = fn();
+	rettime = ktime_get();
+	delta = ktime_sub(rettime, calltime);
+	duration = (unsigned long long) ktime_to_ns(delta) >> 10;
+	pr_debug("initcall %pF returned %d after %lld usecs\n",
+		 fn, ret, duration);
+
+	return ret;
+}
+
+int __init_or_module do_one_initcall(initcall_t fn)
 {
 	int count = preempt_count();
-	ktime_t calltime, delta, rettime;
+	int ret;
 
-	if (initcall_debug) {
-		call.caller = task_pid_nr(current);
-		printk("calling  %pF @ %i\n", fn, call.caller);
-		calltime = ktime_get();
-		trace_boot_call(&call, fn);
-		enable_boot_trace();
-	}
-
-	ret.result = fn();
-
-	if (initcall_debug) {
-		disable_boot_trace();
-		rettime = ktime_get();
-		delta = ktime_sub(rettime, calltime);
-		ret.duration = (unsigned long long) ktime_to_ns(delta) >> 10;
-		trace_boot_ret(&ret, fn);
-		printk("initcall %pF returned %d after %Ld usecs\n", fn,
-			ret.result, ret.duration);
-	}
+	if (initcall_debug)
+		ret = do_one_initcall_debug(fn);
+	else
+		ret = fn();
 
 	msgbuf[0] = 0;
 
-	if (ret.result && ret.result != -ENODEV && initcall_debug)
-		sprintf(msgbuf, "error code %d ", ret.result);
-
 	if (preempt_count() != count) {
-		strlcat(msgbuf, "preemption imbalance ", sizeof(msgbuf));
+		sprintf(msgbuf, "preemption imbalance ");
 		preempt_count() = count;
 	}
 	if (irqs_disabled()) {
 		strlcat(msgbuf, "disabled interrupts ", sizeof(msgbuf));
 		local_irq_enable();
 	}
-	if (msgbuf[0]) {
-		printk("initcall %pF returned with %s\n", fn, msgbuf);
-	}
+	WARN(msgbuf[0], "initcall %pF returned with %s\n", fn, msgbuf);
 
-	return ret.result;
+	return ret;
 }
 
 
-extern initcall_t __initcall_start[], __initcall_end[], __early_initcall_end[];
+extern initcall_t __initcall_start[];
+extern initcall_t __initcall0_start[];
+extern initcall_t __initcall1_start[];
+extern initcall_t __initcall2_start[];
+extern initcall_t __initcall3_start[];
+extern initcall_t __initcall4_start[];
+extern initcall_t __initcall5_start[];
+extern initcall_t __initcall6_start[];
+extern initcall_t __initcall7_start[];
+extern initcall_t __initcall_end[];
+
+static initcall_t *initcall_levels[] __initdata = {
+	__initcall0_start,
+	__initcall1_start,
+	__initcall2_start,
+	__initcall3_start,
+	__initcall4_start,
+	__initcall5_start,
+	__initcall6_start,
+	__initcall7_start,
+	__initcall_end,
+};
+
+/* Keep these in sync with initcalls in include/linux/init.h */
+static char *initcall_level_names[] __initdata = {
+	"early",
+	"core",
+	"postcore",
+	"arch",
+	"subsys",
+	"fs",
+	"device",
+	"late",
+};
+
+static void __init do_initcall_level(int level)
+{
+	extern const struct kernel_param __start___param[], __stop___param[];
+	initcall_t *fn;
+
+	strcpy(static_command_line, saved_command_line);
+	parse_args(initcall_level_names[level],
+		   static_command_line, __start___param,
+		   __stop___param - __start___param,
+		   level, level,
+		   &repair_env_string);
+
+	for (fn = initcall_levels[level]; fn < initcall_levels[level+1]; fn++)
+		do_one_initcall(*fn);
+}
 
 static void __init do_initcalls(void)
 {
-	initcall_t *fn;
+	int level;
 
-	for (fn = __early_initcall_end; fn < __initcall_end; fn++)
-		do_one_initcall(*fn);
-
-	/* Make sure there is no pending stuff from the initcall sequence */
-	flush_scheduled_work();
+	for (level = 0; level < ARRAY_SIZE(initcall_levels) - 1; level++)
+		do_initcall_level(level);
 }
 
 /*
@@ -814,51 +778,62 @@ static void __init do_initcalls(void)
  */
 static void __init do_basic_setup(void)
 {
-	init_workqueues();
 	cpuset_init_smp();
 	usermodehelper_init();
-	init_tmpfs();
+	shmem_init();
 	driver_init();
 	init_irq_proc();
 	do_ctors();
+	usermodehelper_enable();
 	do_initcalls();
+	random_int_secret_init();
 }
 
 static void __init do_pre_smp_initcalls(void)
 {
 	initcall_t *fn;
 
-	for (fn = __initcall_start; fn < __early_initcall_end; fn++)
+	for (fn = __initcall_start; fn < __initcall0_start; fn++)
 		do_one_initcall(*fn);
 }
 
-static void run_init_process(char *init_filename)
+/*
+ * This function requests modules which should be loaded by default and is
+ * called twice right after initrd is mounted and right before init is
+ * exec'd.  If such modules are on either initrd or rootfs, they will be
+ * loaded before control is passed to userland.
+ */
+void __init load_default_modules(void)
 {
-	argv_init[0] = init_filename;
-	kernel_execve(init_filename, argv_init, envp_init);
+	load_default_elevator_module();
 }
 
-/* This is a non __init function. Force it to be noinline otherwise gcc
- * makes it inline to init() and it becomes part of init.text section
- */
-static noinline int init_post(void)
-	__releases(kernel_lock)
+static int run_init_process(const char *init_filename)
 {
+	argv_init[0] = init_filename;
+	return do_execve(init_filename,
+		(const char __user *const __user *)argv_init,
+		(const char __user *const __user *)envp_init);
+}
+
+static noinline void __init kernel_init_freeable(void);
+
+static int __ref kernel_init(void *unused)
+{
+	kernel_init_freeable();
 	/* need to finish all async __init code before freeing the memory */
 	async_synchronize_full();
 	free_initmem();
-	unlock_kernel();
 	mark_rodata_ro();
 	system_state = SYSTEM_RUNNING;
 	numa_default_policy();
 
-
-	current->signal->flags |= SIGNAL_UNKILLABLE;
+	flush_delayed_fput();
 
 	if (ramdisk_execute_command) {
-		run_init_process(ramdisk_execute_command);
-		printk(KERN_WARNING "Failed to execute %s\n",
-				ramdisk_execute_command);
+		if (!run_init_process(ramdisk_execute_command))
+			return 0;
+		pr_err("Failed to execute %s\n", ramdisk_execute_command);
 	}
 
 	/*
@@ -868,251 +843,46 @@ static noinline int init_post(void)
 	 * trying to recover a really broken machine.
 	 */
 	if (execute_command) {
-		run_init_process(execute_command);
-		printk(KERN_WARNING "Failed to execute %s.  Attempting "
-					"defaults...\n", execute_command);
+		if (!run_init_process(execute_command))
+			return 0;
+		pr_err("Failed to execute %s.  Attempting defaults...\n",
+			execute_command);
 	}
-	run_init_process("/sbin/init");
-	run_init_process("/etc/init");
-	run_init_process("/bin/init");
-	run_init_process("/bin/sh");
+	if (!run_init_process("/sbin/init") ||
+	    !run_init_process("/etc/init") ||
+	    !run_init_process("/bin/init") ||
+	    !run_init_process("/bin/sh"))
+		return 0;
 
 	panic("No init found.  Try passing init= option to kernel. "
 	      "See Linux Documentation/init.txt for guidance.");
 }
 
-#ifdef CONFIG_SONOS_FENWAY
-/* bit 22-23 configure the */
-#define SICRH_PWM_FIXED_MASK     (3 << (30 - 10)) /* GPIO pin 5 */
-#define SICRH_PWM_VARIABLE_MASK  (3 << (30 - 22)) /* GPIO pin 11 */
-
-/* bits are 01 for PWM, 11 for GPIO*/
-#define SICRH_PWM_FIXED_PWM      (1 << (30 - 10))
-#define SICRH_PWM_FIXED_GPIO     (3 << (30 - 10))
-#define SICRH_PWM_VARIABLE_PWM   (1 << (30 - 22))
-#define SICRH_PWM_VARIABLE_GPIO  (3 << (30 - 22))
-
-struct timer_list ledtimer;
-int actl_ledtimerdisable=0;
-unsigned char actl_leddata[8]={0,LED_WHITE,0,LED_WHITE,0,LED_WHITE,0,LED_WHITE};
-int actl_ledcur=0;
-
-void audioctl_set_leds(unsigned long arg);
-void audioctl_ledtimer(unsigned long x)
-{
-	if (actl_ledtimerdisable==0)
-		audioctl_set_leds(actl_leddata[actl_ledcur]);
-	actl_ledcur=((actl_ledcur+1)&7);
-	mod_timer(&(ledtimer),jiffies+((3*HZ)/4)); /*750ms into the future*/
-}
-
-/* The white LED pins are multiplexed with timers so that we are able to do
- * PWM dimming. There is currently no requirement to actually do PWM dimming, so
- * the pins are locked into GPIO mode. See the Amoeba HDD for details on the
- * PWM hardware configuration, and @193101 for the code that was removed.
- */
-static void audioctl_configureWhiteLedPins(void)
-{
-	int sicrl_mask, sicrl_bits;
-	int sicrh_mask, sicrh_bits;
-
-	/* we don't need to change the SICRL */
-	sicrl_mask = sicrl_bits = 0;
-
-	sicrh_mask = SICRH_PWM_FIXED_MASK | SICRH_PWM_VARIABLE_MASK;
-	sicrh_bits = SICRH_PWM_FIXED_GPIO | SICRH_PWM_VARIABLE_GPIO;
-	mpc83xx_set_sicr(sicrl_mask, sicrl_bits, sicrh_mask, sicrh_bits);
-}
-
-void audioctl_set_leds(unsigned long arg)
-{
-   u32 gpio;
-   u32 gpio_led_green_mask = 0;
-   u32 gpio_led_red_mask;
-   u32 gpio_led_amber_mask;
-   u32 gpio_led_white_mask;
-
-   gpio = mpc83xx_read_gpio(MPC8315_GPIO_GPDAT);
-
-   if (MDP_SUBMODEL_IS_ANVIL(fenway_submodel)) {
-      gpio_led_white_mask = ANVIL_GPIO_LED_WHITE_MASK;
-      gpio_led_green_mask = ANVIL_GPIO_LED_GREEN_MASK;
-      gpio_led_amber_mask = ANVIL_GPIO_LED_AMBER_MASK;
-      gpio_led_red_mask   = ANVIL_GPIO_LED_RED_MASK;
-      gpio &= ~ANVIL_GPIO_LED_WHITE_FIXED_MASK;
-   }
-   else if (MDP_SUBMODEL_IS_AMOEBA(fenway_submodel)) {
-      gpio_led_white_mask = AMOEBA_GPIO_LED_WHITE_MASK;
-      gpio_led_green_mask = AMOEBA_GPIO_LED_GREEN_MASK;
-      gpio_led_amber_mask = AMOEBA_GPIO_LED_AMBER_MASK;
-      gpio_led_red_mask   = AMOEBA_GPIO_LED_RED_MASK;
-      gpio &= ~ANVIL_GPIO_LED_WHITE_FIXED_MASK;
-   }
-   else {
-      gpio_led_white_mask = FENWAY_GPIO_LED_WHITE_MASK;
-      gpio_led_amber_mask = FENWAY_GPIO_LED_AMBER_MASK;
-      if (sys_mdp.mdp_hwfeatures & MDP_HWFEATURE_LED_SWAP) {
-          /* swapped units with red under play/pause button */
-          gpio_led_red_mask   = FENWAY_GPIO_LED_GREEN_MASK;
-          gpio_led_green_mask = FENWAY_GPIO_LED_RED_MASK;
-      } else {
-          /* unswapped units with green under mute button */
-          gpio_led_red_mask   = FENWAY_GPIO_LED_RED_MASK;
-          gpio_led_green_mask = FENWAY_GPIO_LED_GREEN_MASK;
-      }
-   }
-
-   actl_ledtimerdisable=((arg&LED_OVERRIDE)?1:0);
-
-   /* Fenway family LEDs are active low */
-   if (arg & LED_WHITE)
-      gpio &= ~gpio_led_white_mask;
-   else
-      gpio |= gpio_led_white_mask;
-
-   if (arg & LED_RED)
-      gpio &= ~gpio_led_red_mask;
-   else
-      gpio |= gpio_led_red_mask;
-
-   if (arg & (LED_GREEN | LED_GREEN_BUTTON))
-      gpio &= ~gpio_led_green_mask;
-   else
-      gpio |= gpio_led_green_mask;
-
-   if (arg & LED_AMBER)
-      gpio &= ~gpio_led_amber_mask;
-   else
-      gpio |= gpio_led_amber_mask;
-
-   mpc83xx_write_gpio(MPC8315_GPIO_GPDAT, gpio);
-}
-EXPORT_SYMBOL(audioctl_set_leds);
-
-void audioctl_update_led_data(unsigned char *led_data)
-{
-	memcpy(actl_leddata, led_data, 8);
-
-	/* Force blinking to restart at slot 0  (new pattern) */
-	if (led_data[0] & 0x80) {
-		actl_leddata[0] &= ~0x80;
-		actl_ledcur=0;
-		audioctl_ledtimer(0);
-	}
-}
-EXPORT_SYMBOL(audioctl_update_led_data);
-#endif	// CONFIG_SONOS_FENWAY
-
-#if defined(CONFIG_SONOS_LIMELIGHT)
-#define WHITE_LED_PIN  6
-#define WHITE_LED_MASK (1 << (31 - WHITE_LED_PIN))
-#define IMMR_BASE       0xFFE00000
-#define PMUXCR1_OFFSET  (0x000E0060>>2)
-#define PMUXCR1_SPI     0x00000030
-#define PMUXCR1_GPIO69  0x00000020
-#define GPDIR_OFFSET    (0x0000F000>>2)
-#define GPDAT_OFFSET    (0x0000F008>>2)
-char audioctl_fpga_init_done=0;
-struct timer_list ledtimer;
-__be32 __iomem *pimmr;
-
-void audioctl_notify_fpga_init_done(void)
-{
-   audioctl_fpga_init_done = 1;
-}
-EXPORT_SYMBOL(audioctl_notify_fpga_init_done);
-
-void audioctl_ledtimer(unsigned long x)
-{
-   if (!audioctl_fpga_init_done) {
-      /* toggle the white LED pin */
-      u32 gpdat;
-      gpdat = in_be32(pimmr + GPDAT_OFFSET);
-      gpdat ^= WHITE_LED_MASK;
-      out_be32(pimmr + GPDAT_OFFSET, gpdat);
-      mod_timer(&(ledtimer),jiffies+((3*HZ)/4)); /*750ms into the future*/
-   }
-}
-#endif	// CONFIG_SONOS_LIMELIGHT
-
-static int __init kernel_init(void * unused)
+static noinline void __init kernel_init_freeable(void)
 {
 	/*
 	 * Wait until kthreadd is all set-up.
 	 */
 	wait_for_completion(&kthreadd_done);
-	lock_kernel();
 
-#if defined(CONFIG_SONOS_LIMELIGHT) && !defined(CONFIG_SONOS_DIAGS)
-   /* Get the white LED blinking as soon as possible. The Limelight white LED
-      is controlled by the FPGA which is accessed via the SPI. We can't access
-      the FPGA until the SPI driver is initialized which is much later. The
-      FPGA has a default mode which connects the SPI_MOSI pin to the white LED
-      pin. The SPI_MOSI pin is configured as a GPIO output and is toggled
-      periodically by the timer routine. Once the SPI driver is initialized
-      and the FPGA LED register reconfigured, the pin is configured for SPI by
-      the FPGA driver.
-    */
-   {
-      u32 gpdir, pmuxcr1;
-      pimmr = ioremap(IMMR_BASE, 0xF0000);
-      gpdir = in_be32(pimmr + GPDIR_OFFSET);
-      gpdir |= WHITE_LED_MASK;
-      out_be32(pimmr + GPDIR_OFFSET, gpdir);
-      pmuxcr1 = in_be32(pimmr + PMUXCR1_OFFSET);
-      pmuxcr1 &= ~PMUXCR1_SPI;
-      pmuxcr1 |= PMUXCR1_GPIO69;
-      out_be32(pimmr + PMUXCR1_OFFSET, pmuxcr1);
-
-      init_timer(&ledtimer);
-      ledtimer.data=0;
-      ledtimer.function=audioctl_ledtimer;
-      mod_timer(&(ledtimer),jiffies+((3*HZ)/4)); /*750ms into the future*/
-      //audioctl_ledtimer(0);
-   }
-#endif	// CONFIG_SONOS_LIMELIGHT && !CONFIG_SONOS_DIAGS
+	/* Now the scheduler is fully set up and can do blocking allocations */
+	gfp_allowed_mask = __GFP_BITS_MASK;
 
 	/*
 	 * init can allocate pages on any node
 	 */
-	set_mems_allowed(node_states[N_HIGH_MEMORY]);
+	set_mems_allowed(node_states[N_MEMORY]);
 	/*
 	 * init can run on any cpu.
 	 */
-
-#ifdef CONFIG_SONOS_FENWAY
-	printk("Fenway/Anvil/Amoeba White LED init\n");
-	fenway_gpio_init();
-	if (MDP_SUBMODEL_IS_ANVIL(fenway_submodel)|| MDP_SUBMODEL_IS_AMOEBA(fenway_submodel)) {
-		audioctl_configureWhiteLedPins();
-	}
-
-#ifndef CONFIG_SONOS_DIAGS
-	init_timer(&ledtimer);
-	ledtimer.data=0;
-	ledtimer.function=audioctl_ledtimer;
-	audioctl_ledtimer(0);
-#endif  // CONFIG_SONOS_DIAGS
-
-#endif	// CONFIG_SONOS_FENWAY
-
 	set_cpus_allowed_ptr(current, cpu_all_mask);
-	/*
-	 * Tell the world that we're going to be the grim
-	 * reaper of innocent orphaned children.
-	 *
-	 * We don't want people to have to make incorrect
-	 * assumptions about where in the task array this
-	 * can be found.
-	 */
-	init_pid_ns.child_reaper = current;
 
 	cad_pid = task_pid(current);
 
 	smp_prepare_cpus(setup_max_cpus);
 
 	do_pre_smp_initcalls();
-	start_boot_trace();
+	lockup_detector_init();
 
 	smp_init();
 	sched_init_smp();
@@ -1121,7 +891,7 @@ static int __init kernel_init(void * unused)
 
 	/* Open the /dev/console on the rootfs, this should never fail */
 	if (sys_open((const char __user *) "/dev/console", O_RDWR, 0) < 0)
-		printk(KERN_WARNING "Warning: unable to open an initial console.\n");
+		pr_err("Warning: unable to open an initial console.\n");
 
 	(void) sys_dup(0);
 	(void) sys_dup(0);
@@ -1138,12 +908,20 @@ static int __init kernel_init(void * unused)
 		prepare_namespace();
 	}
 
+#ifdef CONFIG_SONOS_SECBOOT
+	/*
+	 * Check the authorization for sysrq, since when we really want it, we'll
+	 * be in an interrupt and unable to...
+	 */
+	check_sysrq_authorization();
+#endif
+
 	/*
 	 * Ok, we have completed the initial bootup, and
 	 * we're essentially up and running. Get rid of the
 	 * initmem segments and start the user-mode stuff..
 	 */
 
-	init_post();
-	return 0;
+	/* rootfs is available now, try loading default modules */
+	load_default_modules();
 }
