@@ -16,10 +16,18 @@
 #include "br_private.h"
 #include "br_private_stp.h"
 
+#if defined(CONFIG_SONOS)
+#include "br_stp_sonos.h"
+#endif
+
 /* since time values in bpdu are in jiffies and then scaled (1/256)
- * before sending, make sure that is at least one.
+ * before sending, make sure that is at least one STP tick.
  */
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 #define MESSAGE_AGE_INCR	((HZ < 256) ? 1 : (HZ/256))
+#else
+#define MESSAGE_AGE_INCR	((HZ / 256) + 1)
+#endif
 
 static const char *const br_port_state_names[] = {
 	[BR_STATE_DISABLED] = "disabled",
@@ -31,9 +39,15 @@ static const char *const br_port_state_names[] = {
 
 void br_log_state(const struct net_bridge_port *p)
 {
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	br_info(p->br, "port %u(%s) entering %s state\n",
-		(unsigned) p->port_no, p->dev->name,
+		(unsigned int) p->port_no, p->dev->name,
 		br_port_state_names[p->state]);
+#else
+	br_info(p->br, "port %u(%s) entered %s state\n",
+		(unsigned int) p->port_no, p->dev->name,
+		br_port_state_names[p->state]);
+#endif
 }
 
 /* called under bridge lock */
@@ -100,6 +114,23 @@ static int br_should_become_root_port(const struct net_bridge_port *p,
 	return 0;
 }
 
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651, SWPBL-65580 */
+static void br_root_port_block(const struct net_bridge *br,
+			       struct net_bridge_port *p)
+{
+
+	br_notice(br, "port %u(%s) tried to become root port (blocked)",
+		  (unsigned int) p->port_no, p->dev->name);
+
+	p->state = BR_STATE_LISTENING;
+	br_log_state(p);
+	br_ifinfo_notify(RTM_NEWLINK, p);
+
+	if (br->forward_delay > 0)
+		mod_timer(&p->forward_delay_timer, jiffies + br->forward_delay);
+}
+#endif
+
 /* called under bridge lock */
 static void br_root_selection(struct net_bridge *br)
 {
@@ -107,9 +138,16 @@ static void br_root_selection(struct net_bridge *br)
 	u16 root_port = 0;
 
 	list_for_each_entry(p, &br->port_list, list) {
-		if (br_should_become_root_port(p, root_port))
-			root_port = p->port_no;
+		if (!br_should_become_root_port(p, root_port))
+			continue;
 
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651, SWPBL-65580: Use selected root
+			   * port, even if its BR_ROOT_BLOCK flag is set. */
+		if (p->flags & BR_ROOT_BLOCK)
+			br_root_port_block(br, p);
+		else
+#endif
+			root_port = p->port_no;
 	}
 
 	br->root_port = root_port;
@@ -145,17 +183,17 @@ void br_transmit_config(struct net_bridge_port *p)
 	struct br_config_bpdu bpdu;
 	struct net_bridge *br;
 
-
 	if (timer_pending(&p->hold_timer)) {
 		p->config_pending = 1;
 		return;
 	}
 
 	br = p->br;
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	if (!br->stp_enabled)  {
 		return;
 	}
+#endif
 
 	bpdu.topology_change = br->topology_change;
 	bpdu.topology_change_ack = p->topology_change_ack;
@@ -168,31 +206,12 @@ void br_transmit_config(struct net_bridge_port *p)
 	else {
 		struct net_bridge_port *root
 			= br_get_port(br, br->root_port);
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+		sonos_update_message_age(br, root, &bpdu, MESSAGE_AGE_INCR);
+#else
 		bpdu.message_age = (jiffies - root->designated_age)
 			+ MESSAGE_AGE_INCR;
-
-		/* Yes, this does happen.  message_age_timer.expires is
-		 * generally set a little further in the future than we'd like
-		 * (it rounds to 4 jiffies), so it is possible for the above
-		 * calculation to go negative.  That makes something that
-		 * arrived really early appear to have arrived really late,
-		 * which eventually ends up causing the interface to flap.
-		 */
-		if (bpdu.message_age <= 0) {
-
-			//printk("br: %d, expired (root=%d)?\n",
-			//p->port_no, root->port_no);
-
-			if (timer_pending(&root->message_age_timer)) {
-				//printk("br: %d, fixed\n", p->port_no);
-				bpdu.message_age = MESSAGE_AGE_INCR;
-			} else {
-				bpdu.message_age = br->max_age;
-			}
-		}
-
-		//printk("br: %d, age: %d, %d\n",
-		//p->port_no, bpdu.message_age, br->max_age);
+#endif
 	}
 	bpdu.max_age = br->max_age;
 	bpdu.hello_time = br->hello_time;
@@ -208,28 +227,28 @@ void br_transmit_config(struct net_bridge_port *p)
 }
 
 /* called under bridge lock */
-static inline void br_adjust_message_age_timer(struct net_bridge_port *p,
-					       const struct br_config_bpdu *bpdu)
-{
-	mod_timer(&p->message_age_timer, jiffies
-		  + (p->br->max_age - bpdu->message_age));
-}
-
-/* called under bridge lock */
-static inline void br_record_config_information(struct net_bridge_port *p,
-						const struct br_config_bpdu *bpdu)
+static void br_record_config_information(struct net_bridge_port *p,
+					 const struct br_config_bpdu *bpdu)
 {
 	p->designated_root = bpdu->root;
 	p->designated_cost = bpdu->root_path_cost;
 	p->designated_bridge = bpdu->bridge_id;
 	p->designated_port = bpdu->port_id;
-	p->designated_age = jiffies + bpdu->message_age;
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	p->designated_age = jiffies - bpdu->message_age;
+#endif
 
-	br_adjust_message_age_timer(p, bpdu);
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	mod_timer(&p->message_age_timer, jiffies
+		  + (p->br->max_age - bpdu->message_age));
+#else
+	mod_timer(&p->message_age_timer, jiffies
+		  + (bpdu->max_age - bpdu->message_age));
+#endif
 }
 
 /* called under bridge lock */
-static inline void br_record_config_timeout_values(struct net_bridge *br,
+static void br_record_config_timeout_values(struct net_bridge *br,
 					    const struct br_config_bpdu *bpdu)
 {
 	br->max_age = bpdu->max_age;
@@ -241,7 +260,14 @@ static inline void br_record_config_timeout_values(struct net_bridge *br,
 /* called under bridge lock */
 void br_transmit_tcn(struct net_bridge *br)
 {
-	br_send_tcn_bpdu(br_get_port(br, br->root_port));
+	struct net_bridge_port *p;
+
+	p = br_get_port(br, br->root_port);
+	if (p)
+		br_send_tcn_bpdu(p);
+	else
+		br_notice(br, "root port %u not found for topology notice\n",
+			  br->root_port);
 }
 
 /* called under bridge lock */
@@ -288,7 +314,8 @@ static void br_designated_port_selection(struct net_bridge *br)
 }
 
 /* called under bridge lock */
-static int br_supersedes_port_info(struct net_bridge_port *p, struct br_config_bpdu *bpdu)
+static int br_supersedes_port_info(const struct net_bridge_port *p,
+				   const struct br_config_bpdu *bpdu)
 {
 	int t;
 
@@ -319,7 +346,7 @@ static int br_supersedes_port_info(struct net_bridge_port *p, struct br_config_b
 }
 
 /* called under bridge lock */
-static inline void br_topology_change_acknowledged(struct net_bridge *br)
+static void br_topology_change_acknowledged(struct net_bridge *br)
 {
 	br->topology_change_detected = 0;
 	del_timer(&br->tcn_timer);
@@ -330,6 +357,11 @@ void br_topology_change_detection(struct net_bridge *br)
 {
 	int isroot = br_is_root_bridge(br);
 
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	if (br->stp_enabled != BR_KERNEL_STP)
+		return;
+#endif
+
 	br_info(br, "topology change detected, %s\n",
 		isroot ? "propagating" : "sending tcn bpdu");
 
@@ -358,7 +390,7 @@ void br_config_bpdu_generation(struct net_bridge *br)
 }
 
 /* called under bridge lock */
-static inline void br_reply(struct net_bridge_port *p)
+static void br_reply(struct net_bridge_port *p)
 {
 	br_transmit_config(p);
 }
@@ -366,9 +398,10 @@ static inline void br_reply(struct net_bridge_port *p)
 /* called under bridge lock */
 void br_configuration_update(struct net_bridge *br)
 {
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	if (!br->stp_enabled)
 		return;
-
+#endif
 	br_root_selection(br);
 	br_designated_port_selection(br);
 }
@@ -397,6 +430,10 @@ static void br_make_blocking(struct net_bridge_port *p)
 
 		p->state = BR_STATE_BLOCKING;
 		br_log_state(p);
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+		br_ifinfo_notify(RTM_NEWLINK, p);
+#endif
+
 		del_timer(&p->forward_delay_timer);
 	}
 }
@@ -409,17 +446,28 @@ static void br_make_forwarding(struct net_bridge_port *p)
 	if (p->state != BR_STATE_BLOCKING)
 		return;
 
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	if (br->forward_delay == 0) {
+#else
+	if (br->stp_enabled == BR_NO_STP || br->forward_delay == 0) {
+#endif
 		p->state = BR_STATE_FORWARDING;
 		br_topology_change_detection(br);
 		del_timer(&p->forward_delay_timer);
-	}
-	else if (p->br->stp_enabled)
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	} else if (br->stp_enabled)
+#else
+	} else if (br->stp_enabled == BR_KERNEL_STP)
+#endif
 		p->state = BR_STATE_LISTENING;
 	else
 		p->state = BR_STATE_LEARNING;
 
+	br_multicast_enable_port(p);
 	br_log_state(p);
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	br_ifinfo_notify(RTM_NEWLINK, p);
+#endif
 
 	if (br->forward_delay != 0)
 		mod_timer(&p->forward_delay_timer, jiffies + br->forward_delay);
@@ -429,13 +477,24 @@ static void br_make_forwarding(struct net_bridge_port *p)
 void br_port_state_selection(struct net_bridge *br)
 {
 	struct net_bridge_port *p;
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	/* Don't change port states if userspace is handling STP */
 	if (!br->stp_enabled)
 		return;
+#else
+	unsigned int liveports = 0;
+#endif
 
 	list_for_each_entry(p, &br->port_list, list) {
-		if (p->state != BR_STATE_DISABLED) {
+		if (p->state == BR_STATE_DISABLED)
+			continue;
+
+		/* Don't change port states if userspace is handling STP */
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+		if (1) {
+#else
+		if (br->stp_enabled != BR_USER_STP) {
+#endif
 			if (p->port_no == br->root_port) {
 				p->config_pending = 0;
 				p->topology_change_ack = 0;
@@ -449,18 +508,31 @@ void br_port_state_selection(struct net_bridge *br)
 				br_make_blocking(p);
 			}
 		}
+
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+		if (p->state == BR_STATE_FORWARDING)
+			++liveports;
+#endif
 	}
+
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	if (liveports == 0)
+		netif_carrier_off(br->dev);
+	else
+		netif_carrier_on(br->dev);
+#endif
 }
 
 /* called under bridge lock */
-static inline void br_topology_change_acknowledge(struct net_bridge_port *p)
+static void br_topology_change_acknowledge(struct net_bridge_port *p)
 {
 	p->topology_change_ack = 1;
 	br_transmit_config(p);
 }
 
 /* called under bridge lock */
-void br_received_config_bpdu(struct net_bridge_port *p, struct br_config_bpdu *bpdu)
+void br_received_config_bpdu(struct net_bridge_port *p,
+			     const struct br_config_bpdu *bpdu)
 {
 	struct net_bridge *br;
 	int was_root;
@@ -485,20 +557,13 @@ void br_received_config_bpdu(struct net_bridge_port *p, struct br_config_bpdu *b
 		}
 
 		if (p->port_no == br->root_port) {
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 			unsigned long old_max_age = br->max_age;
-
+#endif
 			br_record_config_timeout_values(br, bpdu);
-
-			/* SONOS: If max_age changed we need to bump up the
-			 *        timer again.
-			 */
-			if (br->max_age != old_max_age) {
-				printk("br: max age %ld -> %ld\n",
-				       old_max_age, br->max_age);
-				br_adjust_message_age_timer(p, bpdu);;
-			}
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+			sonos_check_max_age_changed(br, p, bpdu, old_max_age);
+#endif
 			br_config_bpdu_generation(br);
 			if (bpdu->topology_change_ack)
 				br_topology_change_acknowledged(br);
@@ -512,10 +577,77 @@ void br_received_config_bpdu(struct net_bridge_port *p, struct br_config_bpdu *b
 void br_received_tcn_bpdu(struct net_bridge_port *p)
 {
 	if (br_is_designated_port(p)) {
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+		br_info(p->br, "received tcn bpdu on port %u(%s)\n",
+			(unsigned int) p->port_no, p->dev->name);
+#else
 		br_info(p->br, "port %u(%s) received tcn bpdu\n",
-			(unsigned) p->port_no, p->dev->name);
+			(unsigned int) p->port_no, p->dev->name);
+#endif
 
 		br_topology_change_detection(p->br);
 		br_topology_change_acknowledge(p);
 	}
 }
+
+/* Change bridge STP parameter */
+int br_set_hello_time(struct net_bridge *br, unsigned long val)
+{
+	unsigned long t = clock_t_to_jiffies(val);
+
+	if (t < BR_MIN_HELLO_TIME || t > BR_MAX_HELLO_TIME)
+		return -ERANGE;
+
+	spin_lock_bh(&br->lock);
+	br->bridge_hello_time = t;
+	if (br_is_root_bridge(br))
+		br->hello_time = br->bridge_hello_time;
+	spin_unlock_bh(&br->lock);
+	return 0;
+}
+
+int br_set_max_age(struct net_bridge *br, unsigned long val)
+{
+	unsigned long t = clock_t_to_jiffies(val);
+
+	if (t < BR_MIN_MAX_AGE || t > BR_MAX_MAX_AGE)
+		return -ERANGE;
+
+	spin_lock_bh(&br->lock);
+	br->bridge_max_age = t;
+	if (br_is_root_bridge(br))
+		br->max_age = br->bridge_max_age;
+	spin_unlock_bh(&br->lock);
+	return 0;
+
+}
+
+void __br_set_forward_delay(struct net_bridge *br, unsigned long t)
+{
+	br->bridge_forward_delay = t;
+	if (br_is_root_bridge(br))
+		br->forward_delay = br->bridge_forward_delay;
+}
+
+int br_set_forward_delay(struct net_bridge *br, unsigned long val)
+{
+	unsigned long t = clock_t_to_jiffies(val);
+	int err = -ERANGE;
+
+	spin_lock_bh(&br->lock);
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-65580 */
+	if (br->stp_enabled &&
+	    (t < BR_MIN_FORWARD_DELAY || t > BR_MAX_FORWARD_DELAY))
+#else
+	if (br->stp_enabled != BR_NO_STP &&
+	    (t < BR_MIN_FORWARD_DELAY || t > BR_MAX_FORWARD_DELAY))
+#endif
+		goto unlock;
+
+	__br_set_forward_delay(br, t);
+	err = 0;
+
+unlock:
+	spin_unlock_bh(&br->lock);
+	return err;
+}
