@@ -12,14 +12,20 @@
  */
 
 #include <linux/kernel.h>
+#include <linux/kmod.h>
 #include <linux/etherdevice.h>
+#include <linux/rtnetlink.h>
 
 #include "br_private.h"
 #include "br_private_stp.h"
 
+#if defined(CONFIG_SONOS)
+#include "br_sonos.h"
+#include "br_mcast.h"
+#endif
 
 /* Port id is composed of priority and port number.
- * NB: least significant bits of priority are dropped to
+ * NB: some bits of priority are dropped to
  *     make room for more ports.
  */
 static inline port_id br_make_port_id(__u8 priority, __u16 port_no)
@@ -28,6 +34,8 @@ static inline port_id br_make_port_id(__u8 priority, __u16 port_no)
 		| (port_no & ((1<<BR_PORT_BITS)-1));
 }
 
+#define BR_MAX_PORT_PRIORITY ((u16)~0 >> BR_PORT_BITS)
+
 /* called under bridge lock */
 void br_init_port(struct net_bridge_port *p)
 {
@@ -36,44 +44,44 @@ void br_init_port(struct net_bridge_port *p)
 	p->state = BR_STATE_BLOCKING;
 	p->topology_change_ack = 0;
 	p->config_pending = 0;
-	p->direct_enabled = 0;
-	memset(&p->direct_addr, 0, 6);
-	p->direct_last_stp_time = jiffies - BR_DIRECT_STP_TIME;
-	if (p->br->stp_enabled) {
-		br_stp_port_timer_init(p);
-	}
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	sonos_init_port(p);
+#endif
 }
 
-/* called under bridge lock */
+/* NO locks held */
 void br_stp_enable_bridge(struct net_bridge *br)
 {
 	struct net_bridge_port *p;
 
 	spin_lock_bh(&br->lock);
 	mod_timer(&br->hello_timer, jiffies + br->hello_time);
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	mod_timer(&br->gc_timer, jiffies + HZ * 4);
 	mod_timer(&br->mcast_timer, jiffies + br->mcast_advertise_time);
-	
+#else
+	mod_timer(&br->gc_timer, jiffies + HZ/10);
+#endif
+
 	br_config_bpdu_generation(br);
 
 	/* bring up STP-aware ports */
 	list_for_each_entry(p, &br->port_list, list) {
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 		int bUp      = (p->dev->flags & IFF_UP) ? 1 : 0;
 		int bCarrier = netif_carrier_ok(p->dev);
 
-		if (bUp && bCarrier) {
+		if (bUp && bCarrier)
 			br_stp_enable_port(p);
-		}
+#else
+		if (netif_running(p->dev) && netif_oper_up(p->dev))
+			br_stp_enable_port(p);
+#endif
+
 	}
-
-	/* bring up leaf ports */
-	list_for_each_entry(p, &br->leaf_list, list) {
-		if (p->dev->flags & IFF_UP)
-			p->state = BR_STATE_FORWARDING;
-	}
-
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	sonos_enable_leaf_ports(br, p);
+#endif
 	spin_unlock_bh(&br->lock);
 }
 
@@ -82,7 +90,7 @@ void br_stp_disable_bridge(struct net_bridge *br)
 {
 	struct net_bridge_port *p;
 
-	spin_lock(&br->lock);
+	spin_lock_bh(&br->lock);
 
 	/* STP-aware ports */
 	list_for_each_entry(p, &br->port_list, list) {
@@ -91,24 +99,22 @@ void br_stp_disable_bridge(struct net_bridge *br)
 
 	}
 
-	/* Leaf nodes */
-	list_for_each_entry(p, &br->leaf_list, list) {
-		if (p->dev->flags & IFF_UP)
-			p->state = BR_STATE_DISABLED;
-	}
-
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	sonos_disable_leaf_ports(br, p);
 	br_mcast_destroy_list(br);
+#endif
 
 	br->topology_change = 0;
 	br->topology_change_detected = 0;
-
-	spin_unlock(&br->lock);
+	spin_unlock_bh(&br->lock);
 
 	del_timer_sync(&br->hello_timer);
 	del_timer_sync(&br->topology_change_timer);
 	del_timer_sync(&br->tcn_timer);
- 	del_timer_sync(&br->gc_timer);
+	del_timer_sync(&br->gc_timer);
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	del_timer_sync(&br->mcast_timer);
+#endif
 }
 
 /* called under bridge lock */
@@ -116,17 +122,22 @@ void br_stp_enable_port(struct net_bridge_port *p)
 {
 	br_init_port(p);
 	br_port_state_selection(p->br);
+	br_log_state(p);
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	br_ifinfo_notify(RTM_NEWLINK, p);
+#endif
 }
 
 /* called under bridge lock */
 void br_stp_disable_port(struct net_bridge_port *p)
 {
-	struct net_bridge *br;
+	struct net_bridge *br = p->br;
 	int wasroot;
 
-	br = p->br;
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	printk(KERN_INFO "%s: port %i(%s) entering %s state\n",
 	       br->dev->name, p->port_no, p->dev->name, "disabled");
+#endif
 
 	wasroot = br_is_root_bridge(br);
 	br_become_designated_port(p);
@@ -134,10 +145,20 @@ void br_stp_disable_port(struct net_bridge_port *p)
 	p->topology_change_ack = 0;
 	p->config_pending = 0;
 
+	br_log_state(p);
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	br_ifinfo_notify(RTM_NEWLINK, p);
+#endif
+
 	del_timer(&p->message_age_timer);
 	del_timer(&p->forward_delay_timer);
 	del_timer(&p->hold_timer);
 
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	br_fdb_delete_by_port(br, p, 0);
+	br_multicast_disable_port(p);
+#endif
+
 	br_configuration_update(br);
 
 	br_port_state_selection(br);
@@ -146,11 +167,79 @@ void br_stp_disable_port(struct net_bridge_port *p)
 		br_become_root_bridge(br);
 }
 
-/* called under bridge lock */
-static void br_stp_change_bridge_id(struct net_bridge *br,
-				    const unsigned char *addr)
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+static void br_stp_start(struct net_bridge *br)
 {
-	unsigned char oldaddr[6];
+	int r;
+	char *argv[] = { BR_STP_PROG, br->dev->name, "start", NULL };
+	char *envp[] = { NULL };
+
+	r = call_usermodehelper(BR_STP_PROG, argv, envp, UMH_WAIT_PROC);
+
+	spin_lock_bh(&br->lock);
+
+	if (br->bridge_forward_delay < BR_MIN_FORWARD_DELAY)
+		__br_set_forward_delay(br, BR_MIN_FORWARD_DELAY);
+	else if (br->bridge_forward_delay < BR_MAX_FORWARD_DELAY)
+		__br_set_forward_delay(br, BR_MAX_FORWARD_DELAY);
+
+	if (r == 0) {
+		br->stp_enabled = BR_USER_STP;
+		br_debug(br, "userspace STP started\n");
+	} else {
+		br->stp_enabled = BR_KERNEL_STP;
+		br_debug(br, "using kernel STP\n");
+
+		/* To start timers on any ports left in blocking */
+		br_port_state_selection(br);
+	}
+
+	spin_unlock_bh(&br->lock);
+}
+
+static void br_stp_stop(struct net_bridge *br)
+{
+	int r;
+	char *argv[] = { BR_STP_PROG, br->dev->name, "stop", NULL };
+	char *envp[] = { NULL };
+
+	if (br->stp_enabled == BR_USER_STP) {
+		r = call_usermodehelper(BR_STP_PROG, argv, envp, UMH_WAIT_PROC);
+		br_info(br, "userspace STP stopped, return code %d\n", r);
+
+		/* To start timers on any ports left in blocking */
+		spin_lock_bh(&br->lock);
+		br_port_state_selection(br);
+		spin_unlock_bh(&br->lock);
+	}
+
+	br->stp_enabled = BR_NO_STP;
+}
+#endif
+
+void br_stp_set_enabled(struct net_bridge *br, unsigned long val)
+{
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	br->stp_enabled = val?1:0;
+#else
+	ASSERT_RTNL();
+
+	if (val) {
+		if (br->stp_enabled == BR_NO_STP)
+			br_stp_start(br);
+	} else {
+		if (br->stp_enabled != BR_NO_STP)
+			br_stp_stop(br);
+	}
+#endif
+}
+
+/* called under bridge lock */
+void br_stp_change_bridge_id(struct net_bridge *br, const unsigned char *addr)
+{
+	/* should be aligned on 2 bytes for ether_addr_equal() */
+	unsigned short oldaddr_aligned[ETH_ALEN >> 1];
+	unsigned char *oldaddr = (unsigned char *)oldaddr_aligned;
 	struct net_bridge_port *p;
 	int wasroot;
 
@@ -161,12 +250,11 @@ static void br_stp_change_bridge_id(struct net_bridge *br,
 	memcpy(br->dev->dev_addr, addr, ETH_ALEN);
 
 	list_for_each_entry(p, &br->port_list, list) {
-		if (!compare_ether_addr(p->designated_bridge.addr, oldaddr))
+		if (ether_addr_equal(p->designated_bridge.addr, oldaddr))
 			memcpy(p->designated_bridge.addr, addr, ETH_ALEN);
 
-		if (!compare_ether_addr(p->designated_root.addr, oldaddr))
+		if (ether_addr_equal(p->designated_root.addr, oldaddr))
 			memcpy(p->designated_root.addr, addr, ETH_ALEN);
-
 	}
 
 	br_configuration_update(br);
@@ -175,39 +263,43 @@ static void br_stp_change_bridge_id(struct net_bridge *br,
 		br_become_root_bridge(br);
 }
 
-static const unsigned char br_mac_zero[6];
+/* should be aligned on 2 bytes for ether_addr_equal() */
+static const unsigned short br_mac_zero_aligned[ETH_ALEN >> 1];
 
 /* called under bridge lock */
-void br_stp_recalculate_bridge_id(struct net_bridge *br)
+bool br_stp_recalculate_bridge_id(struct net_bridge *br)
 {
+	const unsigned char *br_mac_zero =
+			(const unsigned char *)br_mac_zero_aligned;
 	const unsigned char *addr = br_mac_zero;
 	struct net_bridge_port *p;
 
+#if defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
 	/* SONOS: Allow setting of bridge MAC without funky heuristics below.
 	 *        We don't always have a valid port in the bridge to use for
 	 *        the MAC, sadly.
 	 */
 	if (br->use_static_mac) {
-
-		if (compare_ether_addr(br->bridge_id.addr, br->static_mac)) {
-			br_stp_change_bridge_id(br, br->static_mac);
-		}
-
-	} else {
-
-
-		list_for_each_entry(p, &br->port_list, list) {
-			/* prefer first added wired ethernet interface for
-			   bridge ID */
-			if (addr == br_mac_zero ||
-			    0 == strncmp(p->dev->name, "eth", 3))
-				addr = p->dev->dev_addr;
-		}
-
-		if (compare_ether_addr(br->bridge_id.addr, addr)) {
-			br_stp_change_bridge_id(br, addr);
-		}
+		return change_bridge_id_static_mac(br);
 	}
+#else
+	/* user has chosen a value so keep it */
+	if (br->dev->addr_assign_type == NET_ADDR_SET)
+		return false;
+#endif
+
+	list_for_each_entry(p, &br->port_list, list) {
+		if (addr == br_mac_zero ||
+		    memcmp(p->dev->dev_addr, addr, ETH_ALEN) < 0)
+			addr = p->dev->dev_addr;
+
+	}
+
+	if (ether_addr_equal(br->bridge_id.addr, addr))
+		return false;	/* no change */
+
+	br_stp_change_bridge_id(br, addr);
+	return true;
 }
 
 /* called under bridge lock */
@@ -236,10 +328,16 @@ void br_stp_set_bridge_priority(struct net_bridge *br, u16 newprio)
 }
 
 /* called under bridge lock */
-void br_stp_set_port_priority(struct net_bridge_port *p, u8 newprio)
+int br_stp_set_port_priority(struct net_bridge_port *p, unsigned long newprio)
 {
-	port_id new_port_id = br_make_port_id(newprio, p->port_no);
+	port_id new_port_id;
 
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	if (newprio > BR_MAX_PORT_PRIORITY)
+		return -ERANGE;
+#endif
+
+	new_port_id = br_make_port_id(newprio, p->port_no);
 	if (br_is_designated_port(p))
 		p->designated_port = new_port_id;
 
@@ -250,14 +348,24 @@ void br_stp_set_port_priority(struct net_bridge_port *p, u8 newprio)
 		br_become_designated_port(p);
 		br_port_state_selection(p->br);
 	}
+
+	return 0;
 }
 
 /* called under bridge lock */
-void br_stp_set_path_cost(struct net_bridge_port *p, u32 path_cost)
+int br_stp_set_path_cost(struct net_bridge_port *p, unsigned long path_cost)
 {
+#if !defined(CONFIG_SONOS) /* SONOS SWPBL-19651 */
+	if (path_cost < BR_MIN_PATH_COST ||
+	    path_cost > BR_MAX_PATH_COST)
+		return -ERANGE;
+
+	p->flags |= BR_ADMIN_COST;
+#endif
 	p->path_cost = path_cost;
 	br_configuration_update(p->br);
 	br_port_state_selection(p->br);
+	return 0;
 }
 
 ssize_t br_show_bridge_id(char *buf, const struct bridge_id *id)
@@ -267,114 +375,3 @@ ssize_t br_show_bridge_id(char *buf, const struct bridge_id *id)
 	       id->addr[0], id->addr[1], id->addr[2],
 	       id->addr[3], id->addr[4], id->addr[5]);
 }
-
-
-/* called under bridge lock */
-int br_stp_mod_port_addr(struct net_bridge *br,
-                         unsigned char* oldaddr,
-                         unsigned char* newaddr)
-{
-	struct net_bridge_port *p;
-
-	/*
-	 * Find the port w/ the old mac address.  This should be unique.
-	 *
-	 * This could all get entertaining if we also find a p2p port w/ the
-	 * new MAC.  If so, I don't think we want to do anything since we're
-	 * already going to end up with a topology change.
-	 */
-	list_for_each_entry(p, &br->port_list, list) {
-		if (p->is_p2p && !memcmp(&p->p2p_dest_addr, oldaddr, ETH_ALEN)) {
-			printk("br: Moving %d from "
-			       "%02x:%02x:%02x:%02x:%02x:%02x "
-			       "to %02x:%02x:%02x:%02x:%02x:%02x\n",
-			       p->port_no,
-			       oldaddr[0], oldaddr[1], oldaddr[2],
-			       oldaddr[3], oldaddr[4], oldaddr[5],
-			       newaddr[0], newaddr[1], newaddr[2],
-			       newaddr[3], newaddr[4], newaddr[5]);
-			memcpy(&p->p2p_dest_addr, newaddr, ETH_ALEN);
-			return 0;
-		}
-	}
-
-	return -EINVAL;
-}
-
-/* called under bridge lock */
-int br_stp_mod_port_dev(struct net_bridge *br,
-			unsigned char* oldaddr,
-			struct net_device* dev)
-{
-	struct net_bridge_port *p;
-
-	/*
-	 * Find the port w/ the old mac address.  This should be unique.
-	 */
-	list_for_each_entry(p, &br->port_list, list) {
-
-            if (p->is_p2p && !memcmp(&p->p2p_dest_addr, oldaddr, ETH_ALEN)) {
-			struct net_device *old_dev = p->dev;
-			struct net_bridge_port_list_node* pl_curr  = p->dev->br_port_list;
-			struct net_bridge_port_list_node* pl_trail = NULL;
-
-
-			if (p->dev == dev)
-				return 0;
-
-
-			dev_hold(dev);
-			dev_hold(old_dev);
-
-			printk("br: Moving %d from %s to %s\n",
-				   p->port_no,
-				   old_dev->name,
-				   dev->name);
-
-			/* find the entry in the original dev's list */
-			while (pl_curr) {
-					if (pl_curr->port == p) {
-							break;
-					}
-					pl_trail = pl_curr;
-					pl_curr  = pl_curr->next;
-			}
-
-			if (pl_curr) {
-					/* Remove from one dev list */
-					if (pl_trail) {
-							pl_trail->next = pl_curr->next;
-					} else {
-							old_dev->br_port_list = pl_curr->next;
-					}
-
-					/* Add to the other */
-					pl_curr->next      = dev->br_port_list;
-					dev->br_port_list  = pl_curr;
-
-					/* Update the actual port to point to the dev */
-					p->dev = dev;
-
-					/* MCS-1172: _new_p2p_tunnel has
-					 * done a dev_hold on the old
-					 * device, and _del_p2p_tunnel
-					 * will do a dev_put on the new
-					 * device if the p2p link is
-					 * deleted. To account for this,
-					 * we must do a dev_put on the
-					 * old_dev and a dev_hold on the
-					 * new dev.
-					 */
-					dev_put(old_dev);
-					dev_hold(dev);
-			}
-
-			dev_put(old_dev);
-			dev_put(dev);
-
-			return 0;
-		}
-	}
-
-	return -EINVAL;
-}
